## Goals:

- learn to build minimal Docker image with scratch image as a base
- learn to build minimal Docker image with Google's "distroless" images (

## Minimal Docker image with scratch image as a base

What is the size of the smallest Docker image? There is a special "scratch" image that we can use to base our images from. Lets do the simplest Dockerfile with scratch as a base:

    FROM scratch
    CMD ["/hello"]

(No, you can't run this just yet... Why? There's no "hello" program installed in our image, of course!)
Let's build Docker image, we'll name it "just-scratch":

    matjaz@mbp ~/git/docker-from-scratch: docker build -t just-scratch .
    Sending build context to Docker daemon  4.608kB
    Step 1/2 : FROM scratch
     --->
    Step 2/2 : CMD ["/hello"]
     ---> Running in 81ad67091f56
     ---> 23f50b82759b
    Removing intermediate container 81ad67091f56
    Successfully built 23f50b82759b
    Successfully tagged just-scratch:latest

    matjaz@mbp ~/git/docker-from-scratch: docker images just-scratch
    REPOSITORY          TAG                 IMAGE ID            CREATED             SIZE
    just-scratch        latest              23f50b82759b        35 seconds ago      0B

Docker shows that scratch image uses 0 bytes.

To do something useful, we need a program that we can run. Let's use statically compiled and linked hello-world for Intel 64 bit platform from the official Docker library (https://github.com/docker-library/hello-world/tree/master/amd64/hello-world). Copy file to the project folder, make it executable and modify Dockerfile to copy it into the image:

    matjaz@mbp  ~/git/docker-from-scratch$ curl https://raw.githubusercontent.com/docker-library/hello-world/master/amd64/hello-world/hello -o hello
    matjaz@mbp  ~/git/docker-from-scratch$ chmod +x hello
    matjaz@mbp  ~/git/docker-from-scratch$ ls -la
    total 16
    drwxr-xr-x   4 matjaz  staff   128B Nov  5 11:35 .
    drwxr-xr-x  38 matjaz  staff   1.2K Nov  5 11:11 ..
    -rw-r--r--   1 matjaz  staff    40B Nov  5 11:27 Dockerfile
    -rwxr-xr-x   1 matjaz  staff   1.8K Nov  5 11:34 hello

    FROM scratch
    COPY hello /
    CMD ["/hello"]

Build an image, look at it's size and run the program to check that it works correctly:

    matjaz@mbp ~/git/docker-from-scratch: docker build -t from-scratch . --no-cache
    Sending build context to Docker daemon  4.608kB
    Step 1/3 : FROM scratch
     --->
    Step 2/3 : COPY hello /
     ---> c10cede87c77
    Step 3/3 : CMD ["/hello"]
     ---> Running in 1d2dae8a3074
     ---> fb72fcdbdd4a
    Removing intermediate container 1d2dae8a3074
    Successfully built fb72fcdbdd4a
    Successfully tagged from-scratch:latest
    matjaz@mbp ~/git/docker-from-scratch: docker images from-scratch
    REPOSITORY          TAG                 IMAGE ID            CREATED             SIZE
    from-scratch        latest              fb72fcdbdd4a        10 seconds ago      1.85kB

    matjaz@mbp ~/git/docker-from-scratch: docker run from-scratch

## Distroless Docker images

https://github.com/GoogleContainerTools/distroless

"Distroless" images contain only your application and its runtime dependencies. They do not contain package managers, shells or any other programs you would expect to find in a standard Linux distribution (hence the name: Distroless :).

Advantages:

- restricting what's in your runtime container to precisely what's necessary for your app
- reducing the attack surface (less tools/libs/apps inside container==less attack surface)
- improves signal to noise ration for security/vulnerability scanners

Lets create a Hello world in c with distroless image:

Content of hello.c:


    matjaz@mbp  ~/git/hello$ cat hello.c

    int main () {
    puts("Hello, world!");
    return 0;
    }


Content of Dockerfile:

    matjaz@mbp  ~/git/hello$ cat Dockerfile
    
    FROM ubuntu AS build
    RUN apt-get update
    RUN apt-get install -y build-essential
    COPY hello.c /
    RUN make hello
    FROM  gcr.io/distroless/cc
    COPY --from=build /hello /hello
    CMD ["/hello"]

Build the image, tag it as hello:latest and run it:

    matjaz@mbp  ~/git/hello$ docker build -t hello .
    Sending build context to Docker daemon  3.072kB
    Step 1/8 : FROM ubuntu AS build
    ---> ea4c82dcd15a
    Step 2/8 : RUN apt-get update
    ---> Using cache
    ---> 910ed7151e2e
    Step 3/8 : RUN apt-get install -y build-essential
    ---> Using cache
    ---> 25997a9279f2
    Step 4/8 : COPY hello.c /
    ---> Using cache
    ---> 79984f5e6dbb
    Step 5/8 : RUN make hello
    ---> Using cache
    ---> 1b45b9d1cbba
    Step 6/8 : FROM  gcr.io/distroless/cc
    ---> 9a3e25a372f1
    Step 7/8 : COPY --from=build /hello /hello
    ---> Using cache
    ---> 83f99b1ab872
    Step 8/8 : CMD ["/hello"]
    ---> Using cache
    ---> 009f49961062
    Successfully built 009f49961062
    Successfully tagged hello:latest


    matjaz@mbp  ~/git/hello$ docker run hello
    Hello, world!


Check the image sizes:

    matjaz@mbp  ~/git/hello$ docker images
    REPOSITORY                  TAG                 IMAGE ID            CREATED              SIZE
    hello                       latest              009f49961062        About a minute ago   18.5MB
    <none>                      <none>              1b45b9d1cbba        About a minute ago   315MB
    ubuntu                      16.04               4a689991aa24        4 weeks ago          116MB
    ubuntu                      18.04               ea4c82dcd15a        4 weeks ago          85.8MB
    ubuntu                      latest              ea4c82dcd15a        4 weeks ago          85.8MB
    nginx                       latest              dbfc48660aeb        4 weeks ago          109MB
    busybox                     latest              59788edf1f3e        6 weeks ago          1.15MB
    alpine                      latest              196d12cf6ab1        2 months ago         4.41MB
    jpetazzo/clock              latest              12068b93616f        3 years ago          2.43MB
    gcr.io/distroless/python3   latest              1709d022036c        48 years ago         50.7MB
    gcr.io/distroless/base      latest              1358ac6604c5        48 years ago         16.6MB
    gcr.io/distroless/cc        latest              9a3e25a372f1        48 years ago         18.5MB


You can see that distroless image is quite small (Yes, Alpine Linux is smaller, but it doesn't use "standard" glibc, but instead relies on musl and has a package manager apk included in the image...).

