## Goals:

Show a Kubernetes development process and 0-downtime upgrades for simple Python Flask app:

- rolling updates
- blue-green deployment
- liveness/readiness checks
- Ingress/Ingress controller
- incorporation of some (not all!) of best practices for building and operating containers in Kubernetes

## Prerequisites

- minikube instance (I used latest minikube with a little more RAM (4GB) and K8s v1.17.0 (currently default version for minikube). I deployed it on VirtualBox (--vm-driver parameter):

```
matjaz@mbp (⎈ |minikube:default) ~/git/k8s-upgrade-demo/v1$ minikube start --memory 4G --vm-driver virtualbox --kubernetes-version v1.17.0
😄  minikube v1.6.2 on Darwin 10.14.6
✨  Selecting 'virtualbox' driver from user configuration (alternates: [hyperkit vmwarefusion])
🔥  Creating virtualbox VM (CPUs=2, Memory=4000MB, Disk=20000MB) ...
🐳  Preparing Kubernetes v1.17.0 on Docker '19.03.5' ...
🚜  Pulling images ...
🚀  Launching Kubernetes ...
⌛  Waiting for cluster to come online ...
🏄  Done! kubectl is now configured to use "minikube"

matjaz@mbp ~$ minikube version
minikube version: v1.6.2
commit: 54f28ac5d3a815d1196cd5d57d707439ee4bb392
matjaz@mbp ~$ kubectl version
Client Version: version.Info{Major:"1", Minor:"17", GitVersion:"v1.17.0" ... Platform:"darwin/amd64"}
Server Version: version.Info{Major:"1", Minor:"17", GitVersion:"v1.17.0" ... Platform:"linux/amd64"}
```

Required enabled "minikube addons enable ingress"

```
matjaz@mbp ~$ minikube addons list
- addon-manager: enabled
- dashboard: enabled
- default-storageclass: enabled
- efk: disabled
- freshpod: disabled
- gvisor: disabled
- helm-tiller: disabled
- ingress: enabled
- ingress-dns: disabled
- logviewer: disabled
- metrics-server: disabled
- nvidia-driver-installer: disabled
- nvidia-gpu-device-plugin: disabled
- registry: disabled
- registry-creds: disabled
- storage-provisioner: enabled
- storage-provisioner-gluster: disabled
```

You may have to wait a bit for ingress controller to be in the running state:

```
    matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get pod --namespace=kube-system
    NAME                                        READY   STATUS    RESTARTS   AGE
    coredns-6955765f44-6hlc4                    1/1     Running   0          4m5s
    coredns-6955765f44-hmt5b                    1/1     Running   0          4m5s
    etcd-minikube                               1/1     Running   0          4m11s
    kube-addon-manager-minikube                 1/1     Running   0          4m11s
    kube-apiserver-minikube                     1/1     Running   0          4m11s
    kube-controller-manager-minikube            1/1     Running   0          4m11s
    kube-proxy-jdrxf                            1/1     Running   0          4m5s
    kube-scheduler-minikube                     1/1     Running   0          4m11s
    nginx-ingress-controller-6fc5bcc8c9-7jvt8   1/1     Running   0          4m1s
    storage-provisioner                         1/1     Running   0          4m2s
```

## Developing a python (Flask) web app manually

### v1: simple Flask app 

Lets create a simple Flask app that will return the name of the server where it runs. App is really basic, a copy of hello world, with no tests and it runs on development Flask server. We'll improve this app with some (but not all recommended steps!) of the stuff that will make it more "cloud native".

We will also deliberately just "hand-waive away" some of Python and Flask best practices for production deployment (again, so that code is minimised... just be aware of that!)


v1/app.py

```python
from flask import Flask

import socket

app = Flask(__name__)

hostname = socket.gethostname()


@app.route('/')
def hello():
    return "Hello from version v1 executed on {}\n".format(hostname)


if __name__ == '__main__':
    app.run(host='0.0.0.0', port=8080)
```

- v1/requirements.txt

```
flask==1.1.1
```

- v1/Dockerfile

```
FROM python:3-alpine

# install pip packages
COPY requirements.txt /app/requirements.txt
RUN pip install --no-cache-dir -r /app/requirements.txt

# copy application files
COPY app.py /app

WORKDIR /app

EXPOSE 8080

CMD ["python", "app.py"]
```

Build a Docker image manually, by using a Docker server inside Minikube's VM (this way we don't need a Docker Registry for images!):

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ minikube docker-env
export DOCKER_TLS_VERIFY="1"
export DOCKER_HOST="tcp://192.168.99.100:2376"
export DOCKER_CERT_PATH="/Users/matjaz/.minikube/certs"
# Run this command to configure your shell:
# eval $(minikube docker-env)
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ eval $(minikube docker-env)
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ docker version
Client: Docker Engine - Community
 Version:           19.03.4
 API version:       1.40
 Go version:        go1.12.10
 Git commit:        9013bf5
 Built:             Thu Oct 17 23:44:48 2019
 OS/Arch:           darwin/amd64
 Experimental:      false

Server: Docker Engine - Community
 Engine:
  Version:          19.03.5
  API version:      1.40 (minimum version 1.12)
  Go version:       go1.12.12
  Git commit:       633a0ea838
  Built:            Wed Nov 13 07:28:45 2019
  OS/Arch:          linux/amd64
  Experimental:     false
 containerd:
  Version:          v1.2.10
  GitCommit:        b34a5c8af56e510852c35414db4c1f4fa6172339
 runc:
  Version:          commit: d736ef14f0288d6993a1845745d6756cfc9ddd5a
  GitCommit:
 docker-init:
  Version:          0.18.0
  GitCommit:        fec3683
```

Docker image that we will now build will be created inside Minikube's VM, so it will be available directly in K8s!

Build image for hello:v1 app:

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ docker build -t hello:v1 .
Sending build context to Docker daemon  23.04kB
Step 1/7 : FROM python:3-alpine
3-alpine: Pulling from library/python
e6b0cf9c0882: Pull complete
da0e9bf0cc60: Pull complete
c9ea274ed700: Pull complete
9cfac6ba3733: Pull complete
f6a933c9cadc: Pull complete
Digest: sha256:23f6b2c7c94054a451965e37932716e36b8630461ed635a666f100c441798110
Status: Downloaded newer image for python:3-alpine
 ---> 4c30403e92a1
Step 2/7 : COPY requirements.txt /app/requirements.txt
 ---> 262b2ec26ff8
Step 3/7 : RUN pip install --no-cache-dir -r /app/requirements.txt
 ---> Running in 53bf3d7f2322
Collecting flask==1.1.1
  Downloading https://files.pythonhosted.org/packages/9b/93/628509b8d5dc749656a9641f4caf13540e2cdec85276964ff8f43bbb1d3b/Flask-1.1.1-py2.py3-none-any.whl (94kB)
Collecting Werkzeug>=0.15
  Downloading https://files.pythonhosted.org/packages/ce/42/3aeda98f96e85fd26180534d36570e4d18108d62ae36f87694b476b83d6f/Werkzeug-0.16.0-py2.py3-none-any.whl (327kB)
Collecting itsdangerous>=0.24
  Downloading https://files.pythonhosted.org/packages/76/ae/44b03b253d6fade317f32c24d100b3b35c2239807046a4c953c7b89fa49e/itsdangerous-1.1.0-py2.py3-none-any.whl
Collecting click>=5.1
  Downloading https://files.pythonhosted.org/packages/fa/37/45185cb5abbc30d7257104c434fe0b07e5a195a6847506c074527aa599ec/Click-7.0-py2.py3-none-any.whl (81kB)
Collecting Jinja2>=2.10.1
  Downloading https://files.pythonhosted.org/packages/65/e0/eb35e762802015cab1ccee04e8a277b03f1d8e53da3ec3106882ec42558b/Jinja2-2.10.3-py2.py3-none-any.whl (125kB)
Collecting MarkupSafe>=0.23
  Downloading https://files.pythonhosted.org/packages/b9/2e/64db92e53b86efccfaea71321f597fa2e1b2bd3853d8ce658568f7a13094/MarkupSafe-1.1.1.tar.gz
Building wheels for collected packages: MarkupSafe
  Building wheel for MarkupSafe (setup.py): started
  Building wheel for MarkupSafe (setup.py): finished with status 'done'
  Created wheel for MarkupSafe: filename=MarkupSafe-1.1.1-cp38-none-any.whl size=12629 sha256=2595fa3e744ecd0847b4fce583b4ab9ace83fc42aede6ddbe4fd21d8e2559658
  Stored in directory: /tmp/pip-ephem-wheel-cache-u80aoj85/wheels/f2/aa/04/0edf07a1b8a5f5f1aed7580fffb69ce8972edc16a505916a77
Successfully built MarkupSafe
Installing collected packages: Werkzeug, itsdangerous, click, MarkupSafe, Jinja2, flask
Successfully installed Jinja2-2.10.3 MarkupSafe-1.1.1 Werkzeug-0.16.0 click-7.0 flask-1.1.1 itsdangerous-1.1.0
Removing intermediate container 53bf3d7f2322
 ---> d321ddbf87f1
Step 4/7 : COPY app.py /app
 ---> 94956be928d4
Step 5/7 : WORKDIR /app
 ---> Running in 1b5d0c47b30e
Removing intermediate container 1b5d0c47b30e
 ---> 80d4d1d48e02
Step 6/7 : EXPOSE 8080
 ---> Running in 0ca20a22942a
Removing intermediate container 0ca20a22942a
 ---> f63fead58f41
Step 7/7 : CMD ["python", "app.py"]
 ---> Running in 672dc3fbdabb
Removing intermediate container 672dc3fbdabb
 ---> 3f04c56312e2
Successfully built 3f04c56312e2
Successfully tagged hello:v1
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ docker image ls hello
REPOSITORY          TAG                 IMAGE ID            CREATED             SIZE
hello               v1                  3f04c56312e2        15 seconds ago      122MB
```

Lets test our app in K8s with kubectl run (yes, it's Deployment generator deprecated - k8s is switching to Docker-like Pod generation by default, but it will do for now, as we'll create our Deployment yamls very soon anyway):

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl run hello --image=hello:v1
kubectl run --generator=deployment/apps.v1 is DEPRECATED and will be removed in a future version. Use kubectl run --generator=run-pod/v1 or kubectl create instead.
deployment.apps/hello created
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get all
NAME                        READY   STATUS    RESTARTS   AGE
pod/hello-fc557d4b5-swjwc   1/1     Running   0          21m

NAME                 TYPE        CLUSTER-IP   EXTERNAL-IP   PORT(S)   AGE
service/kubernetes   ClusterIP   10.96.0.1    <none>        443/TCP   6h6m

NAME                    DESIRED   CURRENT   UP-TO-DATE   AVAILABLE   AGE
deployment.apps/hello   1         1         1            1           21m

NAME                              DESIRED   CURRENT   READY   AGE
replicaset.apps/hello-fc557d4b5   1         1         1       21m
```

We see that deployment "hello" is running, with "hello" ReplicaSet and hello-fc557d4b5-swjwc Pod.

Look at the app log. We see that app is running on port 8080, as we set with Docker command EXPOSE 8080 in our Dockerfile:

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl logs hello-fc557d4b5-swjwc
 * Serving Flask app "app" (lazy loading)
 * Environment: production
   WARNING: Do not use the development server in a production environment.
   Use a production WSGI server instead.
 * Debug mode: off
 * Running on http://0.0.0.0:8080/ (Press CTRL+C to quit)
```

(Yes, it's not a good production practice to run Flask app like this - but to make it more simple, we'll just handwaive this away).

Lets test it! We need to be INSIDE a k8s cluster for that, so we'll just create a new interactive Pod and test from there - we need to send HTTP GET to the port 8080 on Pod's IP, so let's look it up first:

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get pod hello-fc557d4b5-swjwc -o wide
NAME                    READY   STATUS    RESTARTS   AGE   IP           NODE       NOMINATED NODE
hello-fc557d4b5-swjwc   1/1     Running   0          30m   172.17.0.8   minikube   <none>

matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl run -ti --rm alpine --image=alpine
kubectl run --generator=deployment/apps.v1 is DEPRECATED and will be removed in a future version. Use kubectl run --generator=run-pod/v1 or kubectl create instead.
If you don't see a command prompt, try pressing enter.
/ # wget -qO- http://172.17.0.8:8080
Hello from version v1 executed on hello-fc557d4b5-swjwc
/ # exit
Session ended, resume using 'kubectl attach alpine-9d85bf65c-kg8db -c alpine -i -t' command when the pod is running
deployment.apps "alpine" deleted
```

Our app works! 

```
Hello from version v1 executed on hello-fc557d4b5-swjwc
```

Lets expose our Deployment. Remember that we have 4 options for Services?

- ClusterIP (default -  "virtual" IP accessible inside k8s cluster)
- NodePort (expose on all nodes on the same port, 30000-32768 by default)
- LoadBalancer (expose service by using external LB - that must be enabled/configured by a cloud provider. Minikube simulates this with using a  NodePort and  command  minikube service name_of_our_service) 
- ExternalName (DNS CNAME record that points somewhere outside the cluster)

We can try and use LoadBalancer, minikube will expose a nodeport service and we can use the service by using minikube service.

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl expose deployment hello --port=8080 --target-port=8080 --type=LoadBalancer
service/hello exposed
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get service
NAME         TYPE           CLUSTER-IP     EXTERNAL-IP   PORT(S)          AGE
hello        LoadBalancer   10.97.67.109   <pending>     8080:30232/TCP   10s
kubernetes   ClusterIP      10.96.0.1      <none>        443/TCP          6h26m

matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ minikube service hello --url
http://192.168.99.100:30232
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ minikube service hello
Opening kubernetes service default/hello in default browser...
```

minikube service hello opens a default browser in gets a page:

minikube exposes a NodePort service on IP/port 192.168.99.100/30232 (we can get Minikube's IP by executing command minikube ip).

If we were using a "real" K8s deployment in the cloud, our service would get external IP from our cloud provider (and usually we would have to pay for that). IP of our service would be seen in <EXTERNAL-IP>. As we're simulating that with minikube, it will always stay as <pending>.
Scaling our app

Our App is not highly available (we have only one Pod!). OK, K8s is still ensuring that our declared 1 Pod is always running. Let's test that and delete that Pod manually and see what happens::

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get pod
NAME                    READY   STATUS    RESTARTS   AGE
hello-fc557d4b5-swjwc   1/1     Running   0          53m
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl delete pod hello-fc557d4b5-swjwc
pod "hello-fc557d4b5-swjwc" deleted
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get pod
NAME                    READY   STATUS    RESTARTS   AGE
hello-fc557d4b5-vt9cp   1/1     Running   0          41s
```

Did you notice, how long it takes to delete a Pod?

If we want to see what happens under the hood, we can use optional --wait=false parameter and command will return immediately and not wait for execution to actually finish:

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl delete pod hello-fc557d4b5-vt9cp  --wait=false
pod "hello-fc557d4b5-vt9cp" deleted
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get pod
NAME                    READY   STATUS        RESTARTS   AGE
hello-fc557d4b5-ttw4z   1/1     Running       0          2s
hello-fc557d4b5-vt9cp   1/1     Terminating   0          93s
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get pod
NAME                    READY   STATUS        RESTARTS   AGE
hello-fc557d4b5-ttw4z   1/1     Running       0          4s
hello-fc557d4b5-vt9cp   1/1     Terminating   0          95s
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get pod
NAME                    READY   STATUS        RESTARTS   AGE
hello-fc557d4b5-ttw4z   1/1     Running       0          16s
hello-fc557d4b5-vt9cp   1/1     Terminating   0          107s
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get pod
NAME                    READY   STATUS        RESTARTS   AGE
hello-fc557d4b5-ttw4z   1/1     Running       0          17s
hello-fc557d4b5-vt9cp   1/1     Terminating   0          108s
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get pod
NAME                    READY   STATUS        RESTARTS   AGE
hello-fc557d4b5-ttw4z   1/1     Running       0          19s
hello-fc557d4b5-vt9cp   1/1     Terminating   0          110s
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get pod
NAME                    READY   STATUS        RESTARTS   AGE
hello-fc557d4b5-ttw4z   1/1     Running       0          20s
hello-fc557d4b5-vt9cp   1/1     Terminating   0          111s
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get pod
NAME                    READY   STATUS    RESTARTS   AGE
hello-fc557d4b5-ttw4z   1/1     Running   0          35s
```

We can see that K8s will immediately see that there is not enough Pods running and spawn a new Pod (with different ID, of course - you do remember that debate about pet vs cow service model?;)

Lets add a few replicas:

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl scale deployment hello --replicas=3
deployment.extensions/hello scaled
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get deployments
NAME    DESIRED   CURRENT   UP-TO-DATE   AVAILABLE   AGE
hello   3         3         3            3           69m
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get pod -o wide
NAME                    READY   STATUS    RESTARTS   AGE   IP            NODE       NOMINATED NODE
hello-fc557d4b5-qhgb6   1/1     Running   0          23s   172.17.0.10   minikube   <none>
hello-fc557d4b5-rq8qw   1/1     Running   0          23s   172.17.0.9    minikube   <none>
hello-fc557d4b5-ttw4z   1/1     Running   0          10m   172.17.0.8    minikube   <none>
```

If we use our app now, we'll get responses from one of our 3 deployed Pods (in a random fashion, remember the talk we had about Services and kube-proxy default behavior during load balancing):

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ minikube service hello --url
http://192.168.99.100:30232

matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ curl http://192.168.99.100:30232
Hello from version v1 executed on hello-fc557d4b5-qhgb6
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ curl http://192.168.99.100:30232
Hello from version v1 executed on hello-fc557d4b5-rq8qw
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ curl http://192.168.99.100:30232
Hello from version v1 executed on hello-fc557d4b5-ttw4z
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ curl http://192.168.99.100:30232
Hello from version v1 executed on hello-fc557d4b5-qhgb6
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ curl http://192.168.99.100:30232
Hello from version v1 executed on hello-fc557d4b5-ttw4z
```

We can also decrease replicas:

```
matjaz@mbp ~/git/kubernetes-upgrade-demo/v1: kubectl scale deployment hello --replicas=2
deployment "hello" scaled
matjaz@mbp ~/git/kubernetes-upgrade-demo/v1: kubectl get deployment
NAME      DESIRED   CURRENT   UP-TO-DATE   AVAILABLE   AGE
hello     2         2         2            2           1h
matjaz@mbp ~/git/kubernetes-upgrade-demo/v1: kubectl scale deployment hello --replicas=3
deployment "hello" scaled
matjaz@mbp ~/git/kubernetes-upgrade-demo/v1: kubectl get deployment
NAME      DESIRED   CURRENT   UP-TO-DATE   AVAILABLE   AGE
hello     3         3         3            3           1h
```

Up until now we used kubectl commands, but this is not the most common usage for defining apps in k8s. For more advanced cases we need to create k8s YAML files with definitions. 

Create a new definition by exporting an existing one:

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get deployment hello -o yaml --export > hello-deployment.yaml
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ cat hello-deployment.yaml
apiVersion: apps/v1
kind: Deployment
metadata:
  annotations:
    deployment.kubernetes.io/revision: "1"
    kubernetes.io/change-cause: kubectl scale deployment hello --replicas=1 --record=true
  creationTimestamp: null
  generation: 1
  labels:
    run: hello
  name: hello
  selfLink: /apis/extensions/v1beta1/namespaces/default/deployments/hello
spec:
  progressDeadlineSeconds: 600
  replicas: 2
  revisionHistoryLimit: 10
  selector:
    matchLabels:
      run: hello
  strategy:
    rollingUpdate:
      maxSurge: 25%
      maxUnavailable: 25%
    type: RollingUpdate
  template:
    metadata:
      creationTimestamp: null
      labels:
        run: hello
    spec:
      containers:
      - image: hello:v1
        imagePullPolicy: IfNotPresent
        name: hello
        resources: {}
        terminationMessagePath: /dev/termination-log
        terminationMessagePolicy: File
      dnsPolicy: ClusterFirst
      restartPolicy: Always
      schedulerName: default-scheduler
      securityContext: {}
      terminationGracePeriodSeconds: 30
status: {}
```

By using --export k8s deletes some fields, but we can make this definition even more simple, as K8s just adds default values for stuff we don't declare. Let's change it a bit:

- remove annotations
- change label "run: hello" into more commonly used "app: hello" (to be exact, even this is replaced by standard k8s labels - see my k8s ppt files - but to be as short as possible, we'll keep this). In any case don't forget to change it throughout a file (metadata, spec.template.metadata and spec.selector)
- add explicit port nr for container
- etc.

ProTip: It is easier to make this changes in a editor that knows k8s APIs (as it checks values!). eg. you can use JetBrains PyCharm IDEs (with additional Kubernetes plugin). Students can use it for free (you just register via UNI-LJ email).

v1/hello-deployment.yaml

```
apiVersion: apps/v1
kind: Deployment
metadata:
  labels:
    app: hello
  name: hello
spec:
  replicas: 3
  selector:
    matchLabels:
      app: hello
  strategy:
    rollingUpdate:
      maxSurge: 25%
      maxUnavailable: 25%
    type: RollingUpdate
  template:
    metadata:
      labels:
        app: hello
    spec:
      containers:
      - image: hello:v1
        imagePullPolicy: IfNotPresent
        name: hello
        ports:
        - containerPort: 8080
      restartPolicy: Always
      terminationGracePeriodSeconds: 30
```

Similar for Service:

v1/hello-service.yaml

```
apiVersion: v1
kind: Service
metadata:
  labels:
    app: hello
  name: hello
spec:
  externalTrafficPolicy: Cluster
  ports:
  - nodePort: 30232
    port: 8080
    protocol: TCP
    targetPort: 8080
  selector:
    app: hello
  type: LoadBalancer
```

Delete now our app and recreate it from YAML files with kubectl apply -f and test if it works.

```
matjaz@mbp ~/git/kubernetes-upgrade-demo/v1: kubectl delete service,deployment hello
service "hello" deleted
deployment "hello" deleted
```

Wait for objects to be deleted:

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get all
NAME                 TYPE        CLUSTER-IP   EXTERNAL-IP   PORT(S)   AGE
service/kubernetes   ClusterIP   10.96.0.1    <none>        443/TCP   8h
```

Recreate with yaml files:

```
matjaz@mbp ~/git/kubernetes-upgrade-demo/v1: kubectl apply -f hello-deployment.yaml
deployment "hello" created
matjaz@mbp ~/git/kubernetes-upgrade-demo/v1: kubectl apply -f hello-service.yaml
service "hello" created
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get all
NAME                        READY   STATUS    RESTARTS   AGE
pod/hello-5bc8b7f79-b8pnm   1/1     Running   0          16s
pod/hello-5bc8b7f79-fh4ks   1/1     Running   0          16s
pod/hello-5bc8b7f79-h865v   1/1     Running   0          16s

NAME                 TYPE           CLUSTER-IP       EXTERNAL-IP   PORT(S)          AGE
service/hello        LoadBalancer   10.104.199.145   <pending>     8080:30232/TCP   16s
service/kubernetes   ClusterIP      10.96.0.1        <none>        443/TCP          8h

NAME                    DESIRED   CURRENT   UP-TO-DATE   AVAILABLE   AGE
deployment.apps/hello   3         3         3            3           16s

NAME                              DESIRED   CURRENT   READY   AGE
replicaset.apps/hello-5bc8b7f79   3         3         3       16s
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ minikube service hello --url
http://192.168.99.100:30232
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ curl $(minikube service hello --url)
Hello from version v1 executed on hello-5bc8b7f79-h865v
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ curl $(minikube service hello --url)
Hello from version v1 executed on hello-5bc8b7f79-fh4ks
```

Or we can use minikube host (K8s VM) and try to access ClusterIP directly (remember how k8s networking works? hosts and Pods can communicate with each other):

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ minikube ssh curl http://10.104.199.145:8080
Hello from version v1 executed on hello-5bc8b7f79-h865v
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ minikube ssh curl http://10.104.199.145:8080
Hello from version v1 executed on hello-5bc8b7f79-h865v
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ minikube ssh curl http://10.104.199.145:8080
Hello from version v1 executed on hello-5bc8b7f79-fh4ks
```

It works! We can now save this YAML app v1 definition in git as we do for python app code :)

### Using Ingress

K8s object Service is really simple, too simple in fact for our use. Let's try to use Ingress controller and Ingress objects instead (https://kubernetes.io/docs/concepts/services-networking/ingress/, https://github.com/kubernetes/ingress-nginx).

Minikube already have an Ingress controller (ingress-nginx - installed via minikube  addons).

Ingress will be available on minikube's IP, but we still need DNS record for our virtual host name. We could use a real DNS, but for development that's an overkill. We can fake a proper DNS name for our virtual host in at least 2 ways: via /etc/hosts file and via external development friendly special services nip.io or xip.io:

1. /etc/hosts file

add our test dns name into /erc/hosts

    sudo echo <ip address from minikube ip> <application domain> >> /etc/hosts

eg. for hello.mojadomena.local on Linux/Mac:

    sudo echo 192.168.99.100 hello hello.mojadomena.local >> /etc/hosts

try with nslookup hello.mojadomena.local, it must return Minikube's IP (192.168.99.100 in my case)

2. using xip.io ali nip.io

A neat trick is to use a special external service that dynamically returns local private such as  xip.io ali nip.io. If you use a name in the format  foo.192.168.99.103.xip.io, DNS returns private IP in the name:

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ nslookup foo.192.168.99.100.xip.io
Server:		2a00:ee0:d::13
Address:	2a00:ee0:d::13#53

Non-authoritative answer:
Name:	foo.192.168.99.100.xip.io
Address: 192.168.99.100

matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ nslookup boo.192.168.99.100.nip.io
Server:		2a00:ee0:d::13
Address:	2a00:ee0:d::13#53

Non-authoritative answer:
Name:	boo.192.168.99.100.nip.io
Address: 192.168.99.100

```

 (see xip.io in nip.io). 

Create an Ingress object (use YOUR minikube's IP! - run "minikube ip"):

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ minikube ip
192.168.99.100
```

v1/ingress.yaml

```
apiVersion: networking.k8s.io/v1beta1
kind: Ingress
metadata:
  name: hello-ingress
spec:
  rules:
  - host: hello.192.168.99.100.xip.io
    http:
      paths:
      - path: /
        backend:
          serviceName: hello
          servicePort: 8080
```

Create  Ingress:

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl apply -f ingres.yaml
ingress.networking.k8s.io/hello-ingress created
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ kubectl get ingress
NAME            HOSTS                         ADDRESS          PORTS   AGE
hello-ingress   hello.192.168.99.100.xip.io   192.168.99.100   80      8s
```

Lets try to use virtual host hello.192.168.99.100.xip.io (resolves into 192.168.99.100 - your miinkube's IP), on port 80:

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ nslookup hello.192.168.99.100.xip.io
Server:		193.2.1.66
Address:	193.2.1.66#53

Non-authoritative answer:
Name:	hello.192.168.99.100.xip.io
Address: 192.168.99.100

matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ curl hello.192.168.99.100.xip.io
Hello from version v1 executed on hello-5bc8b7f79-fh4ks
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ curl hello.192.168.99.100.xip.io
Hello from version v1 executed on hello-5bc8b7f79-b8pnm
matjaz@mbp  ~/git/k8s-upgrade-demo/v1$ curl hello.192.168.99.100.xip.io
Hello from version v1 executed on hello-5bc8b7f79-h865v
```

It works! :) now thats more like a production deployment!

You can define as many vitual hosts as you like. You can use TLS for HTTPS, if you use LetsEncrypt, you can even get free real certificates with automatic renewal or even automatic registration of certs and new virtual host names (via projects external-dns  and cert-manager)... 

## v2: new version of an app

Create a new version of an app. Well add "processing" (fake one, it simulates some computation with time.sleep():

v2/app.py

```python
from flask import Flask

import socket
import time

app = Flask(__name__)

hostname = socket.gethostname()


@app.route('/')
def hello():
    # make this call slower ("add some background processing")
    time.sleep(1)
    return "Hello from version v2 executed on {}\n".format(hostname)


if __name__ == '__main__':
    app.run(host='0.0.0.0', port=8080)
```

Create new image hello:v2 (note the usage of cached layers created from previous builds, its much faster):

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v2$ docker build -t hello:v2 .
Sending build context to Docker daemon  9.728kB
Step 1/7 : FROM python:3-alpine
 ---> 4c30403e92a1
Step 2/7 : COPY requirements.txt /app/requirements.txt
 ---> Using cache
 ---> 262b2ec26ff8
Step 3/7 : RUN pip install --no-cache-dir -r /app/requirements.txt
 ---> Using cache
 ---> d321ddbf87f1
Step 4/7 : COPY app.py /app
 ---> 1dc9d463d221
Step 5/7 : WORKDIR /app
 ---> Running in 5ee83543eace
Removing intermediate container 5ee83543eace
 ---> 7d978a9b9ed7
Step 6/7 : EXPOSE 8080
 ---> Running in 2ca932918e6d
Removing intermediate container 2ca932918e6d
 ---> b0c5c2f8602f
Step 7/7 : CMD ["python", "app.py"]
 ---> Running in c93ca9942d56
Removing intermediate container c93ca9942d56
 ---> 385f295be432
Successfully built 385f295be432
Successfully tagged hello:v2
matjaz@mbp  ~/git/k8s-upgrade-demo/v2$ docker image ls hello
REPOSITORY          TAG                 IMAGE ID            CREATED             SIZE
hello               v2                  385f295be432        3 minutes ago       122MB
hello               v1                  3f04c56312e2        3 hours ago         122MB
```

### v2: blue-green deployment

Upgrade from v1 to v2 with blue-green depoyment. We'll need to be able to run both versions at the same time, and we just switch load-balancer configuration (Ingress controller here) from old version (v1 - "blue deployment") to new version(v2 - "green deployment"). 

Modify Deployment and Serviceso that it uses new version of the app. As we don't need LoadBalancer Service and Node Ports (we have Ingress controller!), we can also delete a few properties from Service.

v2/hello-deployment-v2.yaml

```
apiVersion: apps/v1
kind: Deployment
metadata:
  labels:
    app: hello-v2
  name: hello-v2
spec:
  replicas: 3
  selector:
    matchLabels:
      app: hello-v2
  strategy:
    rollingUpdate:
      maxSurge: 1
      maxUnavailable: 1
    type: RollingUpdate
  template:
    metadata:
      labels:
        app: hello-v2
    spec:
      containers:
      - image: hello:v2
        imagePullPolicy: IfNotPresent
        name: hello-v2
        ports:
        - containerPort: 8080
      restartPolicy: Always
      terminationGracePeriodSeconds: 30
```

v2/hello-service-v2.yaml

```
apiVersion: v1
kind: Service
metadata:
  labels:
    app: hello-v2
  name: hello-v2
spec:
  ports:
  - port: 8080
    protocol: TCP
    targetPort: 8080
  selector:
    app: hello-v2
  type: ClusterIP
```

Create new Deployment/Service and test it:

```
matjaz@mbp ~/git/kubernetes-upgrade-demo/v2: kubectl apply -f hello-deployment-v2.yaml -f hello-service-v2.yaml
deployment "hello-v2" created
service "hello-v2" created
matjaz@mbp ~/git/kubernetes-upgrade-demo/v2: kubectl get all
NAME              DESIRED   CURRENT   UP-TO-DATE   AVAILABLE   AGE
deploy/hello      3         3         3            3           44m
deploy/hello-v2   3         3         3            3           14s

NAME                     DESIRED   CURRENT   READY     AGE
rs/hello-94d7bc97c       3         3         3         44m
rs/hello-v2-645d6445b5   3         3         3         14s

NAME              DESIRED   CURRENT   UP-TO-DATE   AVAILABLE   AGE
deploy/hello      3         3         3            3           44m
deploy/hello-v2   3         3         3            3           14s

NAME                     DESIRED   CURRENT   READY     AGE
rs/hello-94d7bc97c       3         3         3         44m
rs/hello-v2-645d6445b5   3         3         3         14s

NAME                           READY     STATUS        RESTARTS   AGE
po/hello-94d7bc97c-2j9cg       1/1       Running       0          44m
po/hello-94d7bc97c-4k4pq       1/1       Running       0          44m
po/hello-94d7bc97c-4lr4z       1/1       Running       0          44m
po/hello-v2-645d6445b5-2qc68   1/1       Running       0          14s
po/hello-v2-645d6445b5-4gvjd   1/1       Running       0          14s
po/hello-v2-645d6445b5-xk4qb   1/1       Running       0          14s

NAME             TYPE           CLUSTER-IP       EXTERNAL-IP   PORT(S)          AGE
svc/hello        LoadBalancer   10.109.115.231   <pending>     8080:30593/TCP   44m
svc/hello-v2     ClusterIP      10.111.78.127    <none>        8080/TCP         14s
svc/kubernetes   ClusterIP      10.96.0.1        <none>        443/TCP          9h
```

Test new v2 service via Minikube host (because now we have ONLY ClusterIP, so it's available only inside a cluster!):

```
matjaz@mbp ~/git/kubernetes-upgrade-demo/v2: minikube ssh curl http://10.111.78.127:8080
Hello from version v2 executed on hello-v2-645d6445b5-4gvjd
matjaz@mbp ~/git/kubernetes-upgrade-demo/v2: minikube ssh curl http://10.111.78.127:8080
Hello from version v2 executed on hello-v2-645d6445b5-2qc68
matjaz@mbp ~/git/kubernetes-upgrade-demo/v2: minikube ssh curl http://10.111.78.127:8080
Hello from version v2 executed on hello-v2-645d6445b5-xk4qb
```

We can see that it is 1s slower now (added 1s "processing").

Now we can switch our LB (Ingress controller in our case) from v1 to v2. Modify Ingress object to point to new Service v2:

v2/ingress.yaml

```
apiVersion: networking.k8s.io/v1beta1
kind: Ingress
metadata:
  name: hello-ingress
spec:
  rules:
  - host: hello.192.168.99.103.xip.io
    http:
      paths:
      - path: /
        backend:
          serviceName: hello-v2
          servicePort: 8080
```

```
matjaz@mbp ~/git/kubernetes-upgrade-demo/v2: cat ingress.yaml
apiVersion: networking.k8s.io/v1beta1
kind: Ingress
metadata:
  name: hello-ingress
spec:
  rules:
  - host: hello.192.168.99.103.xip.io
    http:
      paths:
      - path: /
        backend:
          serviceName: hello-v2
          servicePort: 8080
matjaz@mbp ~/git/kubernetes-upgrade-demo/v2: kubectl apply -f ingress.yaml
ingress "hello-ingress" configured
```

Switch to new version is practically instantaneous. Rollback is possible by using old version of Ingress object. If version v2 of our app is ok, we could now delete it and release resources (but dont do it, we'll test a rollback to v1 in next step!).

Test the v2 service first:

```
matjaz@mbp ~/git/kubernetes-upgrade-demo/v2: curl http://hello.192.168.99.103.xip.io
Hello from version v2 executed on hello-v2-645d6445b5-xk4qb
matjaz@mbp ~/git/kubernetes-upgrade-demo/v2: curl http://hello.192.168.99.103.xip.io
Hello from version v2 executed on hello-v2-645d6445b5-2qc68
matjaz@mbp ~/git/kubernetes-upgrade-demo/v2: curl http://hello.192.168.99.103.xip.io
Hello from version v2 executed on hello-v2-645d6445b5-4gvjd
```

## v3: rolling update and readiness/liveness probes

Rollback to our old v1 app again (cd into v1 folder!!):

```
matjaz@mbp ~/git/kubernetes-upgrade-demo/v1: kubectl apply -f ingress.yaml
ingress "hello-ingress" configured
matjaz@mbp ~/git/kubernetes-upgrade-demo/v1: curl http://hello.192.168.99.103.xip.io
Hello from version v1 executed on hello-94d7bc97c-4k4pq
```

Delete v2 from cluster (its crap anyway - too slow ;):

```
matjaz@mbp ~/git/kubernetes-upgrade-demo/v3: kubectl delete deployment,service hello-v2
deployment "hello-v2" deleted
service "hello-v2" deleted
```

We're using our old "quick" v1 again. Lest do an update and a rolling update from v1 to v3, but we'll keep names for Deployment and Services (we'll keep names as "hello").

First, fix the code by making it 50% faster than v2 ;), but sadly, for that to happen, we'll need a longer startup time. It will take more than 10s (yes, again all simulated with magic sleep() calls) for our app to be ready to serve traffic (in reality, this could load a large ML model, or initialize a database, or start a Java VM, or connect to a slow external resource, etc.):

v3/app.py

```python
from flask import Flask

import socket
import time

app = Flask(__name__)

hostname = socket.gethostname()


@app.route('/')
def hello():
    # make this call 50% faster than v2 :) ("add some background processing")
    time.sleep(0.5)
    return "Hello from version v3 executed on {}\n".format(hostname)


if __name__ == '__main__':
    # c3 call is faster than v2, but a pod needs 10s warmup time to start serving traffic!
    # this is simulated with sleep(10), which sleeps for 10s, than starts working
    time.sleep(10)
    app.run(host='0.0.0.0', port=8080)
```

Build image for v3:

```
matjaz@mbp ~/git/kubernetes-upgrade-demo/v3: docker build -t hello:v3 .
Sending build context to Docker daemon   7.68kB
Step 1/7 : FROM python:3-alpine
 ---> 024992483c59
Step 2/7 : COPY requirements.txt /app/requirements.txt
 ---> Using cache
 ---> 2ac81535d373
Step 3/7 : RUN pip install --no-cache-dir -r /app/requirements.txt
 ---> Using cache
 ---> b764113c9970
Step 4/7 : COPY app.py /app
 ---> e67bdbd505b1
Removing intermediate container 5861aec856d7
Step 5/7 : WORKDIR /app
 ---> 9c16a467518d
Removing intermediate container 45a429ee438b
Step 6/7 : EXPOSE 8080
 ---> Running in 48d1173a9913
 ---> f95908148ec7
Removing intermediate container 48d1173a9913
Step 7/7 : CMD python app.py
 ---> Running in e1156246b10d
 ---> dcf8ab815058
Removing intermediate container e1156246b10d
Successfully built dcf8ab815058
Successfully tagged hello:v3
```

For rolling update, we **modify hello-deployment.yaml**:

```
strategy:
  type: RollingUpdate
  rollingUpdate:
    maxSurge: 0
    maxUnavailable: 1
minReadySeconds: 12
```

Parameters are:

- type: [RollingUpdate, Recreate]
- maxUnavailable: max unavailable Pods (number or %).
- maxSurge: max Pods OVER the defined number of replicas suring upgrade (nr or %).

Think about: Rolling update with maxSurge: 0, maxUnavailable: 1. What does it mean? What about values 1 and 1? By default this is set to 25% for both (but once rouded up and once rounded down). What does it mean for 10 and 10?

App now cold starts and is not ready to serve traffic for 10s. We need to take that into account...(https://kubernetes.io/docs/concepts/workloads/pods/pod-lifecycle/#container-probes):

- liveness: if ok, it is alive. If not, restart it.
- readiness: if ok, it can server traffic. If not, don't include it in a resource pool to serve traffic.
- startup: similar to liveness, but used only once, when container is starting up for the first time. After startup succeds, liveness Probe parameters take care of restarting a container. 

We'll add simple tests, delay initially for 15 s (so we're sure that 10s init is ok and processed on our crappy laptop hw ;))

```
readinessProbe:
  httpGet:
    path: /ready
    port: 8080
  initialDelaySeconds: 15
  periodSeconds: 5
livenessProbe:
  httpGet:
    path: /live
    port: 8080
  initialDelaySeconds: 15
  periodSeconds: 5
```

v3/hello-deployment.yaml is now (watch for version label and readiness/liveness probes):

```
apiVersion: apps/v1
kind: Deployment
metadata:
  labels:
    app: hello
  name: hello
spec:
  replicas: 3
  selector:
    matchLabels:
      app: hello
  strategy:
    rollingUpdate:
      maxSurge: 0
      maxUnavailable: 1
    type: RollingUpdate
  minReadySeconds: 15
  template:
    metadata:
      labels:
        app: hello
        version: v3
    spec:
      containers:
      - image: hello:v3
        imagePullPolicy: IfNotPresent
        name: hello
        ports:
        - containerPort: 8080
        readinessProbe:
          httpGet:
            path: /ready
            port: 8080
          initialDelaySeconds: 15
          periodSeconds: 5
        livenessProbe:
          httpGet:
            path: /live
            port: 8080
          initialDelaySeconds: 15
          periodSeconds: 5
      restartPolicy: Always
      terminationGracePeriodSeconds: 30
```

In this case we could use different liveness/readiness probes, but let's add an example implementation to our app just for that:

v3/app.py

```python
from flask import Flask, abort

import socket
import time

app = Flask(__name__)

hostname = socket.gethostname()


@app.route('/')
def hello():
    # make this call 50% faster than v2 :) ("add some background processing")
    time.sleep(0.5)
    return "Hello from version v3 executed on {}\n".format(hostname)


@app.route('/live')
def live():
    return "OK"


@app.route('/ready')
def ready():
    # TODO: check system if it's ready to serve traffic, e.g.
    # result = check_system_if_its_working_ok.ping()

    # let's fake a check
    result = "ok"

    if result:
        return "Yes"
    else:
        abort(500)


if __name__ == '__main__':
    # c3 call is faster than v2, but a pod needs 10s to start serving traffic!
    # thi is simulated with sleep(10), which sleeps for 10s, than starts working
    time.sleep(10)
    app.run(host='0.0.0.0', port=8080)
```


Lets build image again (and reuse v3 tag):

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v3$ docker build -t hello:v3 .
Sending build context to Docker daemon  8.192kB
Step 1/7 : FROM python:3-alpine
 ---> 1a8edcb29ce4
Step 2/7 : COPY requirements.txt /app/requirements.txt
 ---> Using cache
 ---> 9a52af3a0f3e
Step 3/7 : RUN pip install --no-cache-dir -r /app/requirements.txt
 ---> Using cache
 ---> 658e2cace4ff
Step 4/7 : COPY app.py /app
 ---> 1b33e8318a31
Step 5/7 : WORKDIR /app
 ---> Running in a08973cb1778
Removing intermediate container a08973cb1778
 ---> 44e35aa27101
Step 6/7 : EXPOSE 8080
 ---> Running in 1ca37ec2d326
Removing intermediate container 1ca37ec2d326
 ---> 074bc629268e
Step 7/7 : CMD ["python", "app.py"]
 ---> Running in 363615d7d90d
Removing intermediate container 363615d7d90d
 ---> cee969175ed0
Successfully built cee969175ed0
Successfully tagged hello:v3
```

Lets run rolling update now from v1 to v3. To easier see whats happening, you can use more than one terminal

Additional terminal 1:

```
matjaz@mbp ~/git/kubernetes-upgrade-demo: kubectl get deployment hello -o wide --watch
NAME      DESIRED   CURRENT   UP-TO-DATE   AVAILABLE   AGE       CONTAINERS   IMAGES     SELECTOR
hello     3         3         3            3           2h        hello        hello:v1   app=hello
hello     3         3         3            3           2h        hello        hello:v3   app=hello
hello     3         3         3            3           2h        hello        hello:v3   app=hello
hello     3         3         0            3           2h        hello        hello:v3   app=hello
hello     3         3         1            2           2h        hello        hello:v3   app=hello
hello     3         3         1            2           2h        hello        hello:v3   app=hello
hello     3         3         1            3           2h        hello        hello:v3   app=hello
hello     3         2         1            2           2h        hello        hello:v3   app=hello
hello     3         3         2            2           2h        hello        hello:v3   app=hello
hello     3         3         2            2           2h        hello        hello:v3   app=hello
hello     3         3         2            3           2h        hello        hello:v3   app=hello
hello     3         2         2            2           2h        hello        hello:v3   app=hello
hello     3         3         3            2           2h        hello        hello:v3   app=hello
hello     3         3         3            2           2h        hello        hello:v3   app=hello
hello     3         3         3            3           2h        hello        hello:v3   app=hello
^Cmatjaz@mbp ~/git/kubernetes-upgrade-demo:
```

additional terminal 2:

```
matjaz@mbp ~/git/kubernetes-upgrade-demo/v2: kubectl get pod --watch
NAME                        READY     STATUS    RESTARTS   AGE
hello-94d7bc97c-2j9cg       1/1       Running   0          2h
hello-94d7bc97c-4k4pq       1/1       Running   0          2h
hello-94d7bc97c-4lr4z       1/1       Running   0          2h
hello-94d7bc97c-4k4pq       1/1       Terminating   0         2h
hello-66596c7757-dvfg8      0/1       Pending   0         0s
hello-66596c7757-dvfg8      0/1       Pending   0         0s
hello-66596c7757-dvfg8      0/1       ContainerCreating   0         0s
hello-66596c7757-dvfg8      0/1       Running   0         1s
hello-66596c7757-dvfg8      1/1       Running   0         24s
hello-94d7bc97c-4k4pq       0/1       Terminating   0         2h
hello-94d7bc97c-4lr4z       1/1       Terminating   0         2h
hello-66596c7757-czmq4      0/1       Pending   0         0s
hello-66596c7757-czmq4      0/1       Pending   0         0s
hello-66596c7757-czmq4      0/1       ContainerCreating   0         0s
hello-66596c7757-czmq4      0/1       Running   0         1s
hello-94d7bc97c-4k4pq       0/1       Terminating   0         2h
hello-94d7bc97c-4k4pq       0/1       Terminating   0         2h
hello-66596c7757-czmq4      1/1       Running   0         15s
hello-94d7bc97c-2j9cg       1/1       Terminating   0         2h
hello-66596c7757-nbp4c      0/1       Pending   0         0s
hello-66596c7757-nbp4c      0/1       Pending   0         1s
hello-66596c7757-nbp4c      0/1       ContainerCreating   0         1s
hello-94d7bc97c-4lr4z       0/1       Terminating   0         2h
hello-66596c7757-nbp4c      0/1       Running   0         2s
hello-66596c7757-nbp4c      1/1       Running   0         24s
hello-94d7bc97c-2j9cg       0/1       Terminating   0         2h
hello-94d7bc97c-4lr4z       0/1       Terminating   0         2h
hello-94d7bc97c-4lr4z       0/1       Terminating   0         2h
hello-94d7bc97c-2j9cg       0/1       Terminating   0         2h
hello-94d7bc97c-2j9cg       0/1       Terminating   0         
...blabla on the end you'll see it converges to final values...
```

```
matjaz@mbp ~/git/kubernetes-upgrade-demo: kubectl get pod
NAME                     READY     STATUS    RESTARTS   AGE
hello-66596c7757-czmq4   1/1       Running   0          19m
hello-66596c7757-dvfg8   1/1       Running   0          20m
hello-66596c7757-nbp4c   1/1       Running   0          18m
```

In additional terminal 3 you can simulate a live user calling our services in the middle of the upgrade:

```
matjaz@mbp ~: while true; do sleep 1; curl http://hello.192.168.99.103.xip.io;done
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v1 executed on hello-94d7bc97c-t579t
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v1 executed on hello-94d7bc97c-t579t
Hello from version v1 executed on hello-94d7bc97c-t579t
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v1 executed on hello-94d7bc97c-t579t
Hello from version v1 executed on hello-94d7bc97c-t579t
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v3 executed on hello-66596c7757-49vzt
Hello from version v3 executed on hello-66596c7757-49vzt
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v3 executed on hello-66596c7757-49vzt
Hello from version v3 executed on hello-66596c7757-xdmsj
Hello from version v1 executed on hello-94d7bc97c-n9j95
Hello from version v3 executed on hello-66596c7757-49vzt
Hello from version v3 executed on hello-66596c7757-49vzt
... all calls are on v3 now!  ...
```

In our primary terminal window we'll start rolling update and watch what happens in all 3 terminals (best to see them all at once).

```
matjaz@mbp ~/git/kubernetes-upgrade-demo/v3: kubectl apply -f hello-deployment.yaml
deployment "hello" configured
```

## v4: Linux signals and running as PID1 inside a container

Our app currently runs as PID1, and this has many caveats (see https://cloud.google.com/solutions/best-practices-for-building-containers#signal-handling)

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v3$ kubectl exec -ti hello-6fd656b97d-44lsf -- ps aux
PID   USER     TIME  COMMAND
    1 root      0:02 python app.py
  497 root      0:00 ps aux
```

That means that our python app needs to react properly on different Signals, especially SIGTERM which Docker uses to try to gracefully stop a container. If you don't react to this, Docker will just kill process after 10s. Do we have this problem with our Flask app? Lets test it directly by using docker (so we exclude K8s deliberate timeouts such as terminationGracePeriodSeconds):

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v3$ docker run -d --name myapp hello:v3
a790353d19c5cdaa33e9c6af0c36dffda07ee1d4d2c1920794397702932722c4
matjaz@mbp  ~/git/k8s-upgrade-demo/v3$ docker ps -l
CONTAINER ID        IMAGE               COMMAND             CREATED             STATUS              PORTS               NAMES
a790353d19c5        hello:v3            "python app.py"     25 seconds ago      Up 23 seconds       8080/tcp            myapp
matjaz@mbp  ~/git/k8s-upgrade-demo/v3$ docker logs myapp
 * Serving Flask app "app" (lazy loading)
 * Environment: production
   WARNING: Do not use the development server in a production environment.
   Use a production WSGI server instead.
 * Debug mode: off
 * Running on http://0.0.0.0:8080/ (Press CTRL+C to quit)
matjaz@mbp  ~/git/k8s-upgrade-demo/v3$ time docker stop myapp
myapp

real	0m10.291s
user	0m0.041s
sys	0m0.023s
```

Yup, looks like Docker gave up after 10s and just killed our process :( (btw, it's even worse in K8s, as the default timeout is 30s. Remember when we were scaling our v1 app and it took a long time to delete a Pod? Now you know why: k8s was waiting for a Pod to terminate)

We can mend this in different ways, common solutions are using special init process manager and/or adding code for Signals to our app. An alternative solution is also to run all containers in the same process namespace (this basically means that PID1 will be in this case a special "placeholder" Pause container). Lets introduce changes to the code and add an init process manager to our container:

v4/app.py

```python
from flask import Flask, abort

import socket
import time

import signal
import sys

def sigterm_handler(_signo, _stack_frame):
    # TODO: start shutting down cleanly
    # we'll just exit for now
    sys.exit(0)

signal.signal(signal.SIGTERM, sigterm_handler)

app = Flask(__name__)

hostname = socket.gethostname()


@app.route('/')
def hello():
    # make this call 50% faster than v2 :) ("add some background processing")
    time.sleep(0.5)
    return "Hello from version v3 executed on {}\n".format(hostname)


@app.route('/live')
def live():
    return "OK"


@app.route('/ready')
def ready():
    # TODO: check system if it's ready to serve traffic, e.g.
    # result = check_system_if_its_working_ok.ping()

    # let's fake a check
    result = "ok"

    if result:
        return "Yes"
    else:
        abort(500)


if __name__ == '__main__':
    # c3 call is faster than v2, but a pod needs 10s to start serving traffic!
    # thi is simulated with sleep(10), which sleeps for 10s, than starts working
    time.sleep(10)
    app.run(host='0.0.0.0', port=8080)
```

v4/Dockerfile

```
FROM python:3-alpine

# install pip packages
COPY requirements.txt /app/requirements.txt
RUN pip install --no-cache-dir -r /app/requirements.txt

# Add Tini
RUN apk add --no-cache tini
ENTRYPOINT ["/sbin/tini", "--"]

# copy application files
COPY app.py /app

WORKDIR /app

EXPOSE 8080

CMD ["python", "app.py"]
```

Build image v4 and test our app:

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v4$ docker build -t hello:v4 .
Sending build context to Docker daemon  8.704kB
Step 1/9 : FROM python:3-alpine
 ---> 1a8edcb29ce4
Step 2/9 : COPY requirements.txt /app/requirements.txt
 ---> Using cache
 ---> 9a52af3a0f3e
Step 3/9 : RUN pip install --no-cache-dir -r /app/requirements.txt
 ---> Using cache
 ---> 658e2cace4ff
Step 4/9 : RUN apk add --no-cache tini
 ---> Running in 9f95d2352048
fetch http://dl-cdn.alpinelinux.org/alpine/v3.8/main/x86_64/APKINDEX.tar.gz
fetch http://dl-cdn.alpinelinux.org/alpine/v3.8/community/x86_64/APKINDEX.tar.gz
(1/1) Installing tini (0.18.0-r0)
Executing busybox-1.28.4-r2.trigger
OK: 17 MiB in 35 packages
Removing intermediate container 9f95d2352048
 ---> 36a60908c454
Step 5/9 : ENTRYPOINT ["/sbin/tini", "--"]
 ---> Running in e8317c0cef7c
Removing intermediate container e8317c0cef7c
 ---> ad2382028cb2
Step 6/9 : COPY app.py /app
 ---> 29e7dcc36dc9
Step 7/9 : WORKDIR /app
 ---> Running in 927d5431580d
Removing intermediate container 927d5431580d
 ---> a0ebdf519458
Step 8/9 : EXPOSE 8080
 ---> Running in 18ff6d17efd3
Removing intermediate container 18ff6d17efd3
 ---> d4d551baf159
Step 9/9 : CMD ["python", "app.py"]
 ---> Running in 45a03638e04a
Removing intermediate container 45a03638e04a
 ---> 395fd142cf91
Successfully built 395fd142cf91
Successfully tagged hello:v4
```

Image is not much larger

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v4$ docker images hello
REPOSITORY          TAG                 IMAGE ID            CREATED             SIZE
hello               v4                  395fd142cf91        42 seconds ago      88.3MB
hello               v3                  cee969175ed0        About an hour ago   88.1MB
hello               v2                  e22143185fa5        2 hours ago         88.1MB
hello               v1                  49e6fa1bdd45        5 hours ago         88.1MB
```

Lets test it:

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v4$ docker run -d --name myapp hello:v4
2d2117e47bbe337d7874368b9a6407e5663c985f0bd08ef8e7a701b71f088599
matjaz@mbp  ~/git/k8s-upgrade-demo/v4$ docker logs myapp
 * Serving Flask app "app" (lazy loading)
 * Environment: production
   WARNING: Do not use the development server in a production environment.
   Use a production WSGI server instead.
 * Debug mode: off
 * Running on http://0.0.0.0:8080/ (Press CTRL+C to quit)
matjaz@mbp  ~/git/k8s-upgrade-demo/v4$ docker exec -ti myapp ps aux
PID   USER     TIME  COMMAND
    1 root      0:00 /sbin/tini -- python app.py
    6 root      0:00 python app.py
    7 root      0:00 ps aux
```

No longer running as PID1 :) Test with docker stop:

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v4$ time docker stop myapp
myapp

real	0m0.689s
user	0m0.048s
sys	0m0.138s
```

Much better! Also, now our app has a chance to do some cleanup before stopping the container!

## v6: Running as non-root user

It is a good security practice to run your apps as a normal, non-priviledged user. Until now, we've been running our processes inside a container as root. Let's change that. We need to change our Dockerfile to install everything as a nonpriviledged user (called simply "nonroot" in our case ;).

v6/Dockerfile

```
FROM python:3-alpine

# Add Tini
RUN apk add --no-cache tini
ENTRYPOINT ["/sbin/tini", "--"]

RUN adduser -D nonroot
USER nonroot
WORKDIR /home/nonroot

# install pip packages
COPY --chown=nonroot:nonroot requirements.txt requirements.txt
RUN pip install --user --no-cache-dir -r requirements.txt

ENV PATH="/home/nonroot/.local/bin:${PATH}"

# copy application files
COPY --chown=nonroot:nonroot app.py app.py

EXPOSE 8080

CMD ["python", "app.py"]

```

First we add "tini" to the image, as we need to install this packages as root. Then we create a new user (with home folder) called "nonroot" (RUN adduser ....), switch from user root to nonroot (USER nonroot) and change workingdir folder, as new unprivileged user can write only in his home folder (WORKDIR /home/nonroot). Every COPY command now needs to include "--chown=user:group" parameter as COPY command runs as root and without this, nonroot user wouldn't be able to read these files. We also use "--user" parameter with pip to install pip packages in users home folder. We add these folders to the PATH, copy our app files into a container (in a user home folder!) and start app as nonroot. We now run our app as a nonpriviledged user:

```
matjaz@mbp  ~/git/k8s-upgrade-demo/v6$ kubectl exec -ti hello-795c5c547c-l4rhm -- ps aux
PID   USER     TIME  COMMAND
    1 nonroot   0:00 /sbin/tini -- python app.py
    7 nonroot   0:01 python app.py
  567 nonroot   0:00 ps aux
```

Now we are much safer, as an attacker needs to first do a privilege escalation attack to become root in a container, and only if successful can he try to attack a host that runs containers.



